@@ -149,14 +149,14 @@
 //! 
 //! ## More examples
 //! 
-//! https://github.com/alttch/pime/tree/main/examples/
+//! <https://github.com/alttch/pime/tree/main/examples/>
 //! 
 use ini::Ini;
 use log::{debug, error};
 use pyo3::prelude::*;
 use pythonize::{depythonize, pythonize};
 use serde_value::Value;
-use std::collections::BTreeMap;
+use std::collections::{btree_map, BTreeMap};
 use std::fmt;
 use std::sync::atomic::{AtomicU8, Ordering};
 use std::time::Duration;
@@ -170,15 +170,11 @@ pub const STATE_STARTED: u8 = 0xff;
 
 const DATACHANNEL_DEFAULT_BUFFER: usize = 1024;
 
+const PIME_POLL_DELAY: Duration = Duration::from_millis(1);
+
 #[macro_use]
 extern crate lazy_static;
 
-macro_rules! tostr {
-    ($s: expr) => {
-        format!("{}", $s)
-    };
-}
-
 static ENGINE_STATE: AtomicU8 = AtomicU8::new(STATE_STOPPED);
 
 #[derive(Debug, Eq, PartialEq)]
@@ -193,17 +189,16 @@ pub enum ErrorKind {
 
 impl fmt::Display for ErrorKind {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
-        use ErrorKind::*;
         write!(
             f,
             "{}",
             match self {
-                PyException => "Python exception",
-                PackError => "Data pack error",
-                UnpackError => "Data unpack error",
-                ExecError => "Task execution error",
-                InternalError => "Internal error",
-                PySyncEngineStateError => "Engine state error",
+                ErrorKind::PyException => "Python exception",
+                ErrorKind::PackError => "Data pack error",
+                ErrorKind::UnpackError => "Data unpack error",
+                ErrorKind::ExecError => "Task execution error",
+                ErrorKind::InternalError => "Internal error",
+                ErrorKind::PySyncEngineStateError => "Engine state error",
             }
         )
     }
@@ -221,13 +216,13 @@ impl fmt::Display for Error {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         if self.kind == ErrorKind::PyException {
             let mut exc = "Python exception".to_owned();
-            self.exception.as_ref().map(|exception| {
+            if let Some(exception) = self.exception.as_ref() {
                 exc += " ";
                 exc += exception;
                 if !self.message.is_empty() {
-                    exc += ":"
+                    exc += ":";
                 }
-            });
+            };
             write!(f, "{} {}", exc, self.message)
         } else {
             write!(f, "{}: {}", self.kind, self.message)
@@ -237,7 +232,7 @@ impl fmt::Display for Error {
 
 impl From<PyErr> for Error {
     fn from(e: PyErr) -> Error {
-        Error::new(ErrorKind::InternalError, tostr!(e))
+        Error::new_internal(e)
     }
 }
 
@@ -246,21 +241,21 @@ where
     T: std::fmt::Debug,
 {
     fn from(e: tokio::sync::mpsc::error::SendError<T>) -> Error {
-        Error::new_internal(tostr!(e))
+        Error::new_internal(e)
     }
 }
 
 impl From<tokio::sync::TryLockError> for Error {
     fn from(e: tokio::sync::TryLockError) -> Error {
-        Error::new_internal(tostr!(e))
+        Error::new_internal(e)
     }
 }
 
 impl Error {
-    pub fn new(kind: ErrorKind, message: String) -> Self {
+    pub fn new<T: fmt::Display>(kind: ErrorKind, message: T) -> Self {
         Self {
             kind,
-            message: message.to_owned(),
+            message: format!("{}", message),
             exception: None,
             traceback: None,
         }
@@ -274,7 +269,7 @@ impl Error {
         }
     }
 
-    fn new_internal(message: String) -> Self {
+    fn new_internal<T: fmt::Display>(message: T) -> Self {
         Self {
             kind: ErrorKind::PySyncEngineStateError,
             message: format!("CRITICAL: PySyncEngine internal error: {}", message),
@@ -311,22 +306,24 @@ pub struct PyTask {
 }
 
 impl PyTask {
-    pub fn new(command: Value, params: BTreeMap<String, Value>) -> Box<Self> {
-        Box::new(Self {
+    #[must_use]
+    pub fn new(command: Value, params: BTreeMap<String, Value>) -> Self {
+        Self {
             command,
             params,
             need_result: true,
             exclusive: false,
-        })
+        }
     }
 
-    pub fn new0(command: Value) -> Box<Self> {
-        Box::new(Self {
+    #[must_use]
+    pub fn new0(command: Value) -> Self {
+        Self {
             command,
             params: BTreeMap::new(),
             need_result: true,
             exclusive: false,
-        })
+        }
     }
 
     pub fn no_wait(&mut self) {
@@ -340,13 +337,13 @@ impl PyTask {
 }
 
 struct DataChannel {
-    tx: Mutex<mpsc::Sender<(u64, Option<Box<PyTask>>)>>,
-    rx: Mutex<mpsc::Receiver<(u64, Option<Box<PyTask>>)>>,
+    tx: Mutex<mpsc::Sender<(u64, Option<PyTask>)>>,
+    rx: Mutex<mpsc::Receiver<(u64, Option<PyTask>)>>,
 }
 
 impl DataChannel {
     fn new() -> Self {
-        let (tx, rx) = mpsc::channel::<(u64, Option<Box<PyTask>>)>(DATACHANNEL_DEFAULT_BUFFER);
+        let (tx, rx) = mpsc::channel::<(u64, Option<PyTask>)>(DATACHANNEL_DEFAULT_BUFFER);
         Self {
             tx: Mutex::new(tx),
             rx: Mutex::new(rx),
@@ -354,7 +351,7 @@ impl DataChannel {
     }
 
     fn set_buffer(&mut self, buffer: usize) {
-        let (tx, rx) = mpsc::channel::<(u64, Option<Box<PyTask>>)>(buffer);
+        let (tx, rx) = mpsc::channel::<(u64, Option<PyTask>)>(buffer);
         self.tx = Mutex::new(tx);
         self.rx = Mutex::new(rx);
     }
@@ -364,16 +361,14 @@ impl DataChannel {
 struct PyTaskResult {
     task_id: u64,
     ready: triggered::Trigger,
-    result: Option<Box<Value>>,
+    result: Option<Value>,
     error: Option<Error>,
 }
 
 impl PyTaskResult {
+    #[allow(clippy::redundant_closure)]
     fn set_result(&mut self, result: Option<Value>) {
-        match result {
-            Some(v) => self.result = Some(Box::new(v)),
-            None => self.result = None,
-        }
+        self.result = result;
     }
     fn set_error(&mut self, error: Error) {
         self.error = Some(error);
@@ -399,8 +394,7 @@ impl PyTaskCounter {
 }
 
 lazy_static! {
-    static ref PY_RESULTS: RwLock<Box<BTreeMap<u64, PyTaskResult>>> =
-        RwLock::new(Box::new(BTreeMap::new()));
+    static ref PY_RESULTS: RwLock<BTreeMap<u64, PyTaskResult>> = RwLock::new(BTreeMap::new());
     static ref TASK_COUNTER: Mutex<PyTaskCounter> = Mutex::new(PyTaskCounter::new());
     static ref DC: RwLock<DataChannel> = RwLock::new(DataChannel::new());
 }
@@ -439,193 +433,198 @@ macro_rules! log_lost_task {
 
 fn report_error(task_id: u64, error: Error) {
     loop {
-        match PY_RESULTS.try_write() {
-            Ok(mut v) => {
-                let o = match v.get_mut(&task_id) {
-                    Some(x) => x,
-                    None => {
-                        log_lost_task!(task_id);
-                        return;
-                    }
-                };
+        if let Ok(mut v) = PY_RESULTS.try_write() {
+            if let Some(o) = v.get_mut(&task_id) {
                 o.set_error(error);
                 o.ready.trigger();
                 break;
             }
-            Err(_) => {
-                std::thread::sleep(Duration::from_millis(1));
-                continue;
+            log_lost_task!(task_id);
+            return;
+        }
+        std::thread::sleep(PIME_POLL_DELAY);
+        continue;
+    }
+}
+
+#[pyfunction]
+fn report_result(
+    py: Python,
+    task_id: u64,
+    result: Option<Py<PyAny>>,
+    error: Option<(String, String, String)>,
+) {
+    let data: Option<Value> = if let Some(r) = result {
+        match depythonize(r.as_ref(py)) {
+            Ok(v) => v,
+            Err(e) => {
+                report_error(task_id, Error::new(ErrorKind::UnpackError, e));
+                return;
             }
         }
+    } else {
+        None
+    };
+    loop {
+        if let Ok(mut v) = PY_RESULTS.try_write() {
+            if let Some(o) = v.get_mut(&task_id) {
+                o.set_result(data);
+                if let Some(e) = error {
+                    o.set_error(Error::new_py(e));
+                }
+                o.ready.trigger();
+                break;
+            }
+            log_lost_task!(task_id);
+            return;
+        }
+        std::thread::sleep(PIME_POLL_DELAY);
+        continue;
     }
 }
 
 impl<'p> PySyncEngine<'p> {
+    /// # Errors
+    ///
+    /// Will return Err if the Python engine is failed to initialize itself
     pub fn new(py: &'p pyo3::Python) -> Result<Self, Error> {
         PySyncEngine::new_engine(py, None)
     }
+    /// # Errors
+    ///
+    /// Will return Err if the Python engine is failed to initialize itself or venv is
+    /// broken/invalid
     pub fn new_venv(py: &'p pyo3::Python, venv_path: &str) -> Result<Self, Error> {
         PySyncEngine::new_engine(py, Some(venv_path))
     }
+    #[allow(clippy::trivially_copy_pass_by_ref)]
+    #[allow(clippy::too_many_lines)]
     fn new_engine(py: &'p pyo3::Python, venv_path: Option<&str>) -> Result<Self, Error> {
         if ENGINE_STATE.load(Ordering::SeqCst) != STATE_STOPPED {
             return Err(Error::new_online());
         }
-        #[pyfunction]
-        fn report_result(
-            py: Python,
-            task_id: u64,
-            result: Option<Py<PyAny>>,
-            error: Option<(String, String, String)>,
-        ) {
-            let data: Option<Value> = match result {
-                Some(r) => match depythonize(r.as_ref(py)) {
-                    Ok(v) => v,
-                    Err(e) => {
-                        report_error(task_id, Error::new(ErrorKind::UnpackError, tostr!(e)));
-                        return;
-                    }
-                },
-                None => None,
-            };
-            loop {
-                match PY_RESULTS.try_write() {
-                    Ok(mut v) => {
-                        let o = match v.get_mut(&task_id) {
-                            Some(x) => x,
-                            None => {
-                                log_lost_task!(task_id);
-                                return;
-                            }
-                        };
-                        o.set_result(data);
-                        match error {
-                            Some(e) => o.set_error(Error::new_py(e)),
-                            None => {}
-                        }
-                        o.ready.trigger();
-                        break;
-                    }
-                    Err(_) => {
-                        std::thread::sleep(Duration::from_millis(1));
-                        continue;
-                    }
+        if let Some(dir) = venv_path {
+            let cfg = format!("{}/pyvenv.cfg", dir);
+            let ini = match Ini::load_from_file(&cfg) {
+                Ok(v) => v,
+                Err(e) => {
+                    return Err(Error::new(
+                        ErrorKind::InternalError,
+                        format!("Unable to read venv config file {}: {}", cfg, e),
+                    ));
                 }
+            };
+            let ver_info = py.version_info();
+            macro_rules! unwrap_ver_err {
+                ($v: expr) => {
+                    match $v {
+                        Ok(v) => v,
+                        Err(e) => {
+                            return Err(Error::new(
+                                ErrorKind::PyException,
+                                format!("Unable to parse venv version info: {}", e),
+                            ));
+                        }
+                    }
+                };
             }
-        }
-        match venv_path {
-            Some(dir) => {
-                let cfg = format!("{}/pyvenv.cfg", dir);
-                let ini = match Ini::load_from_file(&cfg) {
-                    Ok(v) => v,
-                    Err(e) => {
+            macro_rules! unwrap_ver {
+                ($v: expr) => {
+                    if let Some(v) = $v {
+                        v
+                    } else {
                         return Err(Error::new(
-                            ErrorKind::InternalError,
-                            format!("Unable to read venv config file {}: {}", cfg, e),
+                            ErrorKind::PyException,
+                            "Unable to get venv version info".to_owned(),
                         ));
                     }
                 };
-                let ver_info = py.version_info();
-                macro_rules! unwrap_ver_err {
-                    ($v: expr) => {
-                        match $v {
-                            Ok(v) => v,
-                            Err(e) => {
-                                return Err(Error::new(
-                                    ErrorKind::PyException,
-                                    format!("Unable to parse venv version info: {}", e),
-                                ));
-                            }
-                        }
-                    };
-                }
-                macro_rules! unwrap_ver {
-                    ($v: expr) => {
-                        match $v {
-                            Some(v) => v,
-                            None => {
-                                return Err(Error::new(
-                                    ErrorKind::PyException,
-                                    "Unable to get venv version info".to_owned(),
-                                ));
-                            }
-                        }
-                    };
-                }
-                let venv_ver = unwrap_ver!(ini.general_section().get("version"));
-                let mut s = venv_ver.split(".");
-                let venv_major = unwrap_ver_err!(unwrap_ver!(s.next()).parse::<u8>());
-                let venv_minor = unwrap_ver_err!(unwrap_ver!(s.next()).parse::<u8>());
-                if venv_major != ver_info.major || venv_minor != ver_info.minor {
-                    return Err(Error::new(
-                        ErrorKind::PyException,
-                        format!(
-                            "Unable to activate venv, \
+            }
+            let venv_ver = unwrap_ver!(ini.general_section().get("version"));
+            let mut s = venv_ver.split('.');
+            let venv_major = unwrap_ver_err!(unwrap_ver!(s.next()).parse::<u8>());
+            let venv_minor = unwrap_ver_err!(unwrap_ver!(s.next()).parse::<u8>());
+            if venv_major != ver_info.major || venv_minor != ver_info.minor {
+                return Err(Error::new(
+                    ErrorKind::PyException,
+                    format!(
+                        "Unable to activate venv, \
                             Python library version: {}.{}, venv version: {}. \
                             Please switch the library or rebuild venv",
-                            ver_info.major, ver_info.minor, venv_ver
-                        ),
-                    ));
-                }
-                match ini.general_section().get("include-system-site-packages") {
-                    Some(v) if v == "false" => {
-                        debug!("Removing system-site packages from Python path");
-                        py.run(
-                            "import sys;list(map(lambda x:sys.path.remove(x) \
+                        ver_info.major, ver_info.minor, venv_ver
+                    ),
+                ));
+            }
+            if let Some(v) = ini.general_section().get("include-system-site-packages") {
+                if v == "false" {
+                    debug!("Removing system-site packages from Python path");
+                    py.run(
+                        "import sys;list(map(lambda x:sys.path.remove(x) \
                             if x.endswith('-packages') or '/dist-packages/' in x or \
                             '/site-packages/' in x else False, sys.path.copy()))",
-                            None,
-                            None,
-                        )?;
-                    }
-                    _ => {}
+                        None,
+                        None,
+                    )?;
                 }
-                let import_path = format!(
-                    "{}/lib/python{}.{}/site-packages",
-                    dir, ver_info.major, ver_info.minor
-                );
-                debug!("Adding Python venv import path: {}", import_path);
-                py.run(
-                    &format!("import sys;sys.path.insert(0,'{}')", import_path),
-                    None,
-                    None,
-                )?;
-                //py.run(
-                //"import sys;print(sys.path)",
-                //None,
-                //None,
-                //)?;
             }
-            None => {}
+            let import_path = format!(
+                "{}/lib/python{}.{}/site-packages",
+                dir, ver_info.major, ver_info.minor
+            );
+            debug!("Adding Python venv import path: {}", import_path);
+            py.run(
+                &format!("import sys;sys.path.insert(0,'{}')", import_path),
+                None,
+                None,
+            )?;
+            //py.run(
+            //"import sys;print(sys.path)",
+            //None,
+            //None,
+            //)?;
         }
         let neo = py.import("neotasker.embed")?;
         neo.add_function(wrap_pyfunction!(report_result, neo)?)?;
         Ok(Self { neo })
     }
 
+    /// # Errors
+    ///
+    /// Will return Err if the Python neotasker module failed to add the import path
     pub fn add_import_path(&self, dir: &str) -> Result<(), Error> {
-        match self.neo.call_method1("add_import_path", (dir,)) {
-            Ok(_) => Ok(()),
-            Err(e) => Err(Error::new(ErrorKind::PyException, tostr!(e))),
-        }
+        self.neo.call_method1("add_import_path", (dir,))?;
+        Ok(())
     }
 
+    /// # Errors
+    ///
+    /// Will return Err if the Python neotasker module failed to enable debugging
     pub fn enable_debug(&self) -> Result<(), Error> {
         self.neo.call_method0("set_debug")?;
         Ok(())
     }
 
+    /// # Errors
+    ///
+    /// Will return Err if the Python neotasker module failed to set the poll delay
     pub fn set_poll_delay(&self, delay: f32) -> Result<(), Error> {
         self.neo.call_method1("set_poll_delay", (delay,))?;
         Ok(())
     }
 
+    /// # Errors
+    ///
+    /// Will return Err if the Python neotasker module failed to set the thread pool size
     pub fn set_thread_pool_size(&self, min: u32, max: u32) -> Result<(), Error> {
         self.neo
             .call_method1("set_thread_pool_size", ((min, max),))?;
         Ok(())
     }
 
+    /// # Errors
+    ///
+    /// Will return Err if the engine is already launched or Python neotasker module is failed to
+    /// be intialized
     pub fn launch(&self, py: &'p pyo3::Python, broker: &pyo3::PyAny) -> Result<(), Error> {
         need_offline!();
 
@@ -640,61 +639,43 @@ impl<'p> PySyncEngine<'p> {
         ENGINE_STATE.store(STATE_STARTED, Ordering::SeqCst);
         debug!("PySyncEngine started");
         loop {
-            let (task_id, t) = match py.allow_threads(|| rx.blocking_recv()) {
-                Some(v) => v,
-                None => {
-                    return Err(Error::new_internal("channel broken".to_owned()));
-                }
-            };
-            match t {
-                None => {
+            if let Some((task_id, t)) = py.allow_threads(|| rx.blocking_recv()) {
+                if let Some(task) = t {
+                    let command = match pythonize(*py, &task.command) {
+                        Ok(v) => v,
+                        Err(e) => {
+                            if task_id != 0 {
+                                report_error(task_id, Error::new(ErrorKind::PackError, e));
+                            };
+                            continue;
+                        }
+                    };
+                    let params = match pythonize(*py, &task.params) {
+                        Ok(v) => v,
+                        Err(e) => {
+                            if task_id != 0 {
+                                report_error(task_id, Error::new(ErrorKind::PackError, e));
+                            };
+                            continue;
+                        }
+                    };
+                    if task.exclusive {
+                        if let Err(e) = call_direct.call1((task_id, broker, command, params)) {
+                            report_error(task_id, Error::new(ErrorKind::ExecError, e));
+                        }
+                    } else if task.need_result {
+                        if let Err(e) = call.call1((task_id, broker, command, params)) {
+                            report_error(task_id, Error::new(ErrorKind::ExecError, e));
+                        }
+                    } else {
+                        let _r = spawn.call1((broker, command, params));
+                    }
+                } else {
                     ENGINE_STATE.store(STATE_STOPPING, Ordering::SeqCst);
                     break;
                 }
-                Some(task) => {
-                    let command = pythonize(*py, &task.command);
-                    let params = pythonize(*py, &task.params);
-                    if command.is_err() {
-                        if task_id != 0 {
-                            report_error(
-                                task_id,
-                                Error::new(ErrorKind::PackError, tostr!(params.err().unwrap())),
-                            );
-                        }
-                        continue;
-                    }
-                    if params.is_err() {
-                        if task_id != 0 {
-                            report_error(
-                                task_id,
-                                Error::new(ErrorKind::PackError, tostr!(params.err().unwrap())),
-                            );
-                        }
-                        continue;
-                    }
-                    if task.exclusive {
-                        match call_direct.call1((
-                            task_id,
-                            broker,
-                            command.unwrap(),
-                            params.unwrap(),
-                        )) {
-                            Ok(_) => {}
-                            Err(e) => {
-                                report_error(task_id, Error::new(ErrorKind::ExecError, tostr!(e)));
-                            }
-                        }
-                    } else if task.need_result {
-                        match call.call1((task_id, broker, command.unwrap(), params.unwrap())) {
-                            Ok(_) => {}
-                            Err(e) => {
-                                report_error(task_id, Error::new(ErrorKind::ExecError, tostr!(e)));
-                            }
-                        }
-                    } else {
-                        let _ = spawn.call1((broker, command.unwrap(), params.unwrap()));
-                    }
-                }
+            } else {
+                return Err(Error::new_internal("channel broken".to_owned()));
             }
         }
         debug!("Stopping PySyncEngine");
@@ -705,7 +686,10 @@ impl<'p> PySyncEngine<'p> {
     }
 }
 
-pub async fn call(task: Box<PyTask>) -> Result<Option<Box<Value>>, Error> {
+/// # Errors
+///
+/// Will return Err if the engine is stopped or broken
+pub async fn call(task: PyTask) -> Result<Option<Value>, Error> {
     need_online!();
     if !task.need_result {
         DC.read()
@@ -717,27 +701,20 @@ pub async fn call(task: Box<PyTask>) -> Result<Option<Box<Value>>, Error> {
             .await?;
         return Ok(None);
     }
-    let task_id;
-    loop {
-        let cid = TASK_COUNTER.lock().await.get();
-        match PY_RESULTS.read().await.get(&cid) {
-            Some(_) => critical!("dead tasks in result map"),
-            None => {
-                task_id = cid;
-                break;
-            }
-        };
-    }
     let (trigger, listener) = triggered::trigger();
-    PY_RESULTS.write().await.insert(
-        task_id,
-        PyTaskResult {
-            task_id,
-            result: None,
-            error: None,
-            ready: trigger,
-        },
-    );
+    let task_id = loop {
+        let cid = TASK_COUNTER.lock().await.get();
+        if let btree_map::Entry::Vacant(x) = PY_RESULTS.write().await.entry(cid) {
+            x.insert(PyTaskResult {
+                task_id: cid,
+                result: None,
+                error: None,
+                ready: trigger,
+            });
+            break cid;
+        }
+        critical!("dead tasks in result map");
+    };
     DC.read()
         .await
         .tx
@@ -746,21 +723,20 @@ pub async fn call(task: Box<PyTask>) -> Result<Option<Box<Value>>, Error> {
         .send((task_id, Some(task)))
         .await?;
     listener.await;
-    let res = match PY_RESULTS.write().await.remove(&task_id) {
-        Some(v) => v,
-        None => {
-            return Err(Error::new(
+    PY_RESULTS.write().await.remove(&task_id).map_or_else(
+        || {
+            Err(Error::new(
                 ErrorKind::InternalError,
                 "CRITICAL: Result not found, engine broken".to_owned(),
-            ));
-        }
-    };
-    match res.error {
-        Some(e) => Err(e),
-        None => Ok(res.result),
-    }
+            ))
+        },
+        |res| res.error.map_or(Ok(res.result), Err),
+    )
 }
 
+/// # Errors
+///
+/// Will return Err if the engine is already stopped
 pub async fn stop() -> Result<(), Error> {
     need_online!();
     DC.read().await.tx.lock().await.send((0, None)).await?;
@@ -772,22 +748,26 @@ pub fn get_engine_state() -> u8 {
     ENGINE_STATE.load(Ordering::SeqCst)
 }
 
+#[must_use]
 pub fn is_engine_started() -> bool {
     ENGINE_STATE.load(Ordering::SeqCst) == STATE_STARTED
 }
 
 pub async fn wait_online() {
     while ENGINE_STATE.load(Ordering::SeqCst) != STATE_STARTED {
-        sleep(Duration::from_millis(10)).await;
+        sleep(PIME_POLL_DELAY).await;
     }
 }
 
 pub async fn wait_offline() {
     while ENGINE_STATE.load(Ordering::SeqCst) != STATE_STOPPED {
-        sleep(Duration::from_millis(10)).await;
+        sleep(PIME_POLL_DELAY).await;
     }
 }
 
+/// # Panics
+///
+/// Will panic if the engine is on and the data channel is busy
 pub fn set_mpsc_buffer(buffer: usize) {
     DC.try_write().unwrap().set_buffer(buffer);
 }
